import httplib
import json
import urlparse
import uuid

import pytest
from pytest_localserver.http import WSGIServer
from werkzeug import Request
from werkzeug import Response


class FestungApp(WSGIServer):
    def __init__(self, *args, **kwargs):
        kwargs.update(application=self)
        super(FestungApp, self).__init__(*args, **kwargs)

        self.queries = []
        self.responses = []

    @Request.application
    def __call__(self, request):
        try:
            status, content = self.responses[len(self.queries)]
        except IndexError:
            response = Response("No prepared response")
            response.status_code = httplib.NOT_IMPLEMENTED
            return response

        self.queries.append(request)

        response = Response(json.dumps(content))
        response.status_code = status
        response.headers['content-type'] = 'application/json'

        return response

    def add_response(self, data, status=httplib.OK):
        self.responses.append((status, data))

    @property
    def json_queries(self):
        return [json.loads(q.data) for q in self.queries]


@pytest.fixture
def festung():
    server = FestungApp()
    server.start()
    try:
        yield server
    finally:
        server.stop()


@pytest.fixture
def festung_url(festung):
    return festung.url


@pytest.fixture
def database():
    return uuid.uuid4().hex


@pytest.fixture
def database_url(festung_url, database):
    return urlparse.urljoin(festung_url, database)



import datetime
import decimal

import pytest
import pytz

import festung.dbapi


@pytest.fixture
def connection(database_url):
    conn = festung.dbapi.Connection(database_url)
    try:
        yield conn
    finally:
        conn.close()


@pytest.fixture
def cursor(connection):
    cur = connection.cursor()
    try:
        yield cur
    finally:
        cur.close()


class ResponseFixture(object):
    @pytest.fixture(autouse=True)
    def prepare_response(self, festung):
        festung.add_response(self.response)

    @property
    def data(self):
        return self.response['data']


class TestEmptyResponse(ResponseFixture):
    response = {'data': [], 'headers': []}

    def test_execute(self, festung, cursor):
        query = 'SELECT * FROM foo'
        params = ['string', 1, None]
        cursor.execute(query, params)

        [data] = festung.json_queries
        assert data['sql'] == query
        assert data['params'] == params

    @pytest.mark.parametrize('param,expected_serialialized', [
        (datetime.datetime(1970, 1, 1, 0, 0, tzinfo=pytz.UTC), '1970-01-01T00:00:00+00:00'),
        (datetime.date(1970, 1, 1), '1970-01-01'),
        (datetime.time(0, 0, 0), '00:00:00'),
        (decimal.Decimal('1234.5678'), '1234.5678'),
    ])
    def test_binding_cast(self, festung, cursor, param, expected_serialialized):
        cursor.execute('SELECT 1', [param])
        [data] = festung.json_queries
        [serialized] = data['params']

        assert serialized == expected_serialialized


class TestDummyResponseWithExecutedCursor(ResponseFixture):
    response = {'data': [['a', 1], ['b', 2]], 'headers': ['bar', 'baz']}

    @pytest.fixture(autouse=True)
    def run_query(self, cursor):
        cursor.execute("SELECT * FROM foo")

    def test_fetchone(self, cursor):
        assert cursor.fetchone() == tuple(self.data[0])
        assert cursor.fetchone() == tuple(self.data[1])
        assert cursor.fetchone() is None

    def test_fetchall_same_size_than_response(self, cursor):
        cursor.arraysize = len(self.data)
        assert cursor.fetchmany() == [tuple(row) for row in self.data]

    def test_fetchall_smaller_size_than_response(self, cursor):
        cursor.arraysize = 1
        assert cursor.fetchmany() == [tuple(self.data[0])]

    def test_fetchall_bigger_size_than_response(self, cursor):
        cursor.arraysize = len(self.data) * 10
        assert cursor.fetchmany() == [tuple(row) for row in self.data]



from collections import namedtuple

import requests

from festung._private import cast
from festung._private import no_password_url
from festung import exceptions


__all__ = ['connect', 'apilevel', 'paramstyle', 'threadsafety', 'Connection', 'Cursor']


apilevel = '2.0'
threadsafety = 3  # Threads may share the module, connections and cursors
paramstyle = 'qmark'


class Connection(object):
    def __init__(self, url):
        self.url = url

    def close(self):
        pass

    def commit(self):
        raise NotImplementedError

    def rollback(self):
        raise NotImplementedError

    def cursor(self):
        return Cursor(self)

    def _request(self, method, *args, **kwargs):
        resp = requests.request(method, self.url, *args, **kwargs)
        resp.raise_for_status()
        return resp.json()

    def __repr__(self):
        return "<Connection({})>".format(no_password_url(self.url))


connect = Connection


CursorDescription = namedtuple(
    'CursorDescription', 'name,type_code,display_size,internal_size,precisison,scale,null_ok')


NO_EXECUTE_ROWCOUNT = -1  # <https://www.python.org/dev/peps/pep-0249/#rowcount>
NO_EXECUTE_DESCRIPTION = None  # <https://www.python.org/dev/peps/pep-0249/#description>
NO_MORE_ROW = None  # <https://www.python.org/dev/peps/pep-0249/#fetchone>
NO_EXECUTE_ITER = object()


class Cursor(object):
    def __init__(self, connection):
        self.connection = connection
        self._iter = NO_EXECUTE_ITER
        self._description = NO_EXECUTE_DESCRIPTION
        self._rowcount = NO_EXECUTE_ROWCOUNT
        self._arraysize = 1

    @property
    def description(self):
        raise NotImplementedError
        return self._description

    @property
    def rowcount(self):
        return self._rowcount

    @property
    def arraysize(self):
        return self._arraysize

    @arraysize.setter
    def arraysize(self, value):
        if not isinstance(value, (int, long)):
                raise ValueError("arraysize should be an integer")
        self._arraysize = value

    def callproc(self, procname, parameters=None):
        parameters = parameters or []
        raise NotImplementedError

    def execute(self, operation, parameters=None):
        parameters = parameters or []
        data = dict(sql=operation, params=[cast(p) for p in parameters])
        res = self._request('POST', json=data)
        # TODO(Antoine): Read column description
        self._iter = iter(res['data'])

    def executemany(self, operation, parameters):
        raise NotImplementedError

    def fetchone(self):
        if self._iter is NO_EXECUTE_ITER:
            raise exceptions.ProgrammingError("No statement was executed on this cursor.")
        try:
            return tuple(next(self._iter))
        except StopIteration:
            return NO_MORE_ROW

    def fetchmany(self, size=None):
        if size is None:
            size = self.arraysize

        acc = []
        for _ in range(size):
            row = self.fetchone()
            if row is NO_MORE_ROW:
                break
            acc.append(row)
        return acc

    def nextset(self):
        raise NotImplementedError

    def setinputsize(sizes):
        raise NotImplementedError

    def setoutputsize(size, columns=None):
        raise NotImplementedError

    @property
    def rownumber(self):
        raise NotImplementedError

    def close(self):
        pass

    def _request(self, *args, **kwargs):
        return self.connection._request(*args, **kwargs)



import datetime
import decimal
import StringIO
import urllib
import urlparse
import warnings


__all__ = ['no_password_url']


def _build_netloc(login, password, hostname, port):
    """Build a netloc from login, password, hostname and port. (hostname is required)

    Many librairies try to implement this, they are all doing it wrong:
        * hostname can be an IPv6. `2001:db8::1` needs to built as `[2001:db8::1]`.
        * port can be a string, `http://example.com:http/` is the same as `http://example.com:80/`

    This implementation of building the netloc, supports passing hidden password with stars.
    These passwords won't get url encoded.

    Example:
        >>> _build_netloc(None, None, 'example.com', None)
        'example.com'
        >>> _build_netloc('h@c|<3r', None, 'example.com', None)
        'h%40c%7C%3C3r@example.com'
        >>> _build_netloc('login', '******', '2001:db8::1', 8080)
        'login:******@[2001:db8::1]:8080'
        >>> _build_netloc(None, 'password', 'example.net', None)
        ':password@example.net'
    """
    if not isinstance(hostname, basestring):
        raise ValueError("hostname has to be a string, cannot be none")

    buf = StringIO.StringIO()

    if login is not None:
        buf.write(urllib.quote(login))

    if password is not None:
        buf.write(':')
        buf.write(urllib.quote(password, safe='*'))

    if buf.tell() > 0:
        buf.write('@')

    hostname = urllib.quote(hostname, safe=':')
    if ':' in hostname:
        hostname = '[{}]'.format(hostname)
    buf.write(hostname)

    if port is not None:
        buf.write(':')
        buf.write(urllib.quote(str(port)))

    return buf.getvalue()


def no_password_url(url):
    """Strip the password from the url.

    Example:

        >>> no_password_url('http://login:password@host:8080/path')
        'http://login:********@host:8080/path'
        >>> no_password_url('http://login@[2001:db8::2]/path')
        'http://login@[2001:db8::2]/path'
    """
    parsed_url = urlparse.urlsplit(url)

    password = parsed_url.password
    if password is not None:
        password = '********'  # We don't want to leak the length of the password
    netloc = _build_netloc(parsed_url.username, password, parsed_url.hostname, parsed_url.port)

    return urlparse.urlunsplit((parsed_url.scheme, netloc, parsed_url.path, parsed_url.query,
                               parsed_url.fragment))


def cast(obj):
    """Cast most python types to valid types to be serialized in JSON and passed to SQLite"""
    if obj is None or isinstance(obj, (long, int, float, basestring)):
        if isinstance(obj, float):
            warnings.warn(
                RuntimeWarning, "floats are highly inaccurate. Use them at your own risk.")
        return obj

    if isinstance(obj, decimal.Decimal):
        return str(obj)

    if isinstance(obj, datetime.datetime):
        if obj.tzinfo is None:
            warnings.warn(
                RuntimeWarning, "non-timezone aware datetime can lead to many issues.")
        return obj.isoformat()

    if isinstance(obj, (datetime.date, datetime.time)):
        return obj.isoformat()

    raise TypeError("Unserializeable type {}".format(type(obj)))
